@@ -71,7 +71,7 @@ cr_drpm_create(cr_DeltaTargetPackage *old,
     if (ret != DRPM_ERR_OK) {
         g_set_error(err, ERR_DOMAIN, CRE_DELTARPM,
                     "Deltarpm cannot make %s (%d) from old: %s and new: %s", drpmpath, ret, old->path, new->path);
-        free(drpmpath);
+        g_free(drpmpath);
         drpm_make_options_destroy(&opts);
         return NULL;
     }
@@ -130,6 +130,7 @@ cr_deltapackage_from_drpm_base(const char *filename,
 
     deltapackage->nevr = cr_safe_string_chunk_insert_null(
                                     deltapackage->chunk, str);
+    free(str);
 
     ret = drpm_get_string(delta, DRPM_TAG_SEQUENCE, &str);
     if (ret != DRPM_ERR_OK) {
@@ -141,6 +142,7 @@ cr_deltapackage_from_drpm_base(const char *filename,
 
     deltapackage->sequence = cr_safe_string_chunk_insert_null(
                                     deltapackage->chunk, str);
+    free(str);
 
     drpm_destroy(&delta);
 
@@ -328,7 +330,8 @@ cr_delta_thread(gpointer data, gpointer udata)
             cr_DeltaTargetPackage *old = lelem->data;
 
             g_debug("Generating delta %s -> %s", old->path, tpkg->path);
-            cr_drpm_create(old, tpkg, user_data->outdeltadir, &tmp_err);
+            char * drpm_path = cr_drpm_create(old, tpkg, user_data->outdeltadir, &tmp_err);
+            free(drpm_path);
             if (tmp_err) {
                 g_warning("Cannot generate delta %s -> %s : %s",
                           old->path, tpkg->path, tmp_err->message);
@@ -338,6 +341,8 @@ cr_delta_thread(gpointer data, gpointer udata)
             if (++x == user_data->num_deltas)
                 break;
         }
+
+        g_slist_free_full(local_candidates, (GDestroyNotify) cr_deltatargetpackage_free);
     }
 
     g_debug("Deltas for \"%s\" (%"G_GINT64_FORMAT") generated",
@@ -420,6 +425,7 @@ cr_deltarpms_parallel_deltas(GSList *targetpackages,
                              &tmp_err);
     if (tmp_err) {
         g_propagate_prefixed_error(err, tmp_err, "Cannot create delta pool: ");
+        g_list_free_full(targets, (GDestroyNotify) cr_deltapackage_free);
         return FALSE;
     }
 
@@ -765,8 +771,7 @@ cr_prestodelta_thread(gpointer data, gpointer udata)
         // 1. Remove the key and value from the table without freeing them
         g_hash_table_steal(user_data->ht, key);
         // 2. Append to the list (the value from the hash table)
-        GSList *list = (GSList *) pval;
-        list = g_slist_append(pval, xml_chunk);
+        GSList *list = g_slist_append(pval, xml_chunk);
         // 3. Insert the modified list again
         g_hash_table_insert(user_data->ht, pkey, list);
     } else {
